/*                                     ___   ___               */
/*           ▄▄▄       ██▓     ██▓███  |\  \ |\  \             */
/*          ▒████▄    ▓██▒    ▓██░  ██▒\ \  \\_\  \            */
/*          ▒██  ▀█▄  ▒██░    ▓██░ ██▓▒ \ \______  \           */
/*          ░██▄▄▄▄██ ▒██░    ▒██▄█▓▒ ▒  \|_____|\  \          */
/*           ▓█   ▓██▒░██████▒▒██▒ ░  ░         \ \__\         */
/*           ▒▒   ▓▒█░░ ▒░▓  ░▒▓▒░ ░  ░          \|__|         */
/*            ▒   ▒▒ ░░ ░ ▒  ░░▒ ░                             */
/*         F a░a s▒e && ░S░c h░ä f e r  \                      */
/*                ░  ░    ░  ░            S o  2 0 1 9         */



/*▄▄▄▄▄▄▄▄▄▄▄▄▄▄▄▄▄▄▄▄▄▄▄▄▄▄ INCLUDE ▄▄▄▄▄▄▄▄▄▄▄▄▄▄▄▄▄▄▄▄▄▄▄▄▄▄*/
#include "locks.h"
/*▀▀▀▀▀▀▀▀▀▀▀▀▀▀▀▀▀▀▀▀▀▀▀▀▀▀ INCLUDE ▀▀▀▀▀▀▀▀▀▀▀▀▀▀▀▀▀▀▀▀▀▀▀▀▀▀*/


/*▄▄▄▄▄▄▄▄▄▄▄▄▄▄▄▄▄▄▄▄▄▄ GLOBAL VARIABLES ▄▄▄▄▄▄▄▄▄▄▄▄▄▄▄▄▄▄▄▄▄*/
uint64_t  * ENTERING ;
int64_t  * PRIORITAET;
uint64_t FAVOURED = 0;
uint64_t NUMBER_OF_OBJECTS;          // TODO max 16 @ world
/*▀▀▀▀▀▀▀▀▀▀▀▀▀▀▀▀▀▀▀▀▀▀ GLOBAL VARIABLES ▀▀▀▀▀▀▀▀▀▀▀▀▀▀▀▀▀▀▀▀▀*/


/*▄▄▄▄▄▄▄▄▄▄▄▄▄▄▄▄▄▄▄▄▄▄▄ IMPLEMENTATION ▄▄▄▄▄▄▄▄▄▄▄▄▄▄▄▄▄▄▄▄▄▄*/

/***************************************************************/
/* setup the vaiables for the lock algorithms                  */
void locks_init(uint64_t num)
{
  NUMBER_OF_OBJECTS = num;
  ENTERING   = calloc(NUMBER_OF_OBJECTS, sizeof(uint64_t));
  PRIORITAET = calloc(NUMBER_OF_OBJECTS, sizeof(int64_t));
}

/***************************************************************/
/* remove the memory for the globals                           */
void locks_free(void)
{
  free(ENTERING);
  free(PRIORITAET);
}


/***************************************************************/
/* lock critical section with lamport algorithm                */
void lock_lamport(uint64_t p_id)
{
  /* set the priority of our process  */
  ENTERING[p_id] = 1;
  PRIORITAET[p_id] = 1 + max_of_array(PRIORITAET, NUMBER_OF_OBJECTS);
  ENTERING[p_id] = 0;

  /* look if another proccess has to be processed first        */
  for (size_t i = 0 ; i < NUMBER_OF_OBJECTS; i++)
  {
    while(ENTERING[i])
      ;    // busy waiting until another process retrives priority
    while(PRIORITAET[i] != 0 && PRIORITAET[i] < PRIORITAET[p_id] )
      ;    // busy waiting until another process with lower priority finished
  }
}

/***************************************************************/
/* unlock critical section with lamport algorithm              */
void unlock_lamport(uint64_t p_id)
{
  PRIORITAET[p_id] = 0;

}

/***************************************************************/
/* lock critical section with dekker algorithm                 */
/* this works only for 2 Processes                             */
void lock_dekker(uint64_t p_id)
{
  ENTERING[p_id] = 1;
  while ( ENTERING[NUMBER_OF_OBJECTS - 1 - p_id] ) {            //
    if (FAVOURED != p_id) {
      ENTERING[p_id] = 0;
      while( FAVOURED != p_id )
        ; // busy waiting
      ENTERING[p_id] = 1;
    }
  }
}

/***************************************************************/
/* unlock critical section with dekker algorithm               */
void unlock_dekker(uint64_t p_id)
{
  FAVOURED = NUMBER_OF_OBJECTS - 1 - p_id;
  ENTERING[p_id] = 0;
}

/***************************************************************/
/* lock critical section with peterson algorithm               */
void lock_peterson(uint64_t p_id)
{
  ENTERING[p_id] = 1;
  FAVOURED = NUMBER_OF_OBJECTS - 1 - p_id;
  while ( ENTERING[NUMBER_OF_OBJECTS - 1 - p_id]
       && FAVOURED == (NUMBER_OF_OBJECTS - 1 - p_id) )
    ;            // busy waiting for another thread to finish

}

/***************************************************************/
/* unlock critical section with peterson algorithm             */
void unlock_peterson(uint64_t p_id)
{
  ENTERING[p_id] = 0;
}

/*▀▀▀▀▀▀▀▀▀▀▀▀▀▀▀▀▀▀▀▀▀▀ IMPLEMENTATION ▀▀▀▀▀▀▀▀▀▀▀▀▀▀▀▀▀▀▀▀▀▀▀*/



/*▄▄▄▄▄▄▄▄▄▄▄▄▄▄▄▄▄▄▄▄▄▄ HELPER FUNCTIONS ▄▄▄▄▄▄▄▄▄▄▄▄▄▄▄▄▄▄▄▄▄*/

/***************************************************************/
/* returns array max                                           */
int64_t max_of_array(int64_t * arr, size_t length)
{
  int64_t ret = INT64_MIN;
  for (size_t i = 0; i < length  ; i++)
  {
    if (ret < arr[i] )
      ret = arr[i];
  }
  return ret;
}

/*▀▀▀▀▀▀▀▀▀▀▀▀▀▀▀▀▀▀▀▀▀▀ HELPER FUNCTIONS ▀▀▀▀▀▀▀▀▀▀▀▀▀▀▀▀▀▀▀▀▀▀*/
