/*                                     ___   ___               */
/*           ▄▄▄       ██▓     ██▓███  |\  \ |\  \             */
/*          ▒████▄    ▓██▒    ▓██░  ██▒\ \  \\_\  \            */
/*          ▒██  ▀█▄  ▒██░    ▓██░ ██▓▒ \ \______  \           */
/*          ░██▄▄▄▄██ ▒██░    ▒██▄█▓▒ ▒  \|_____|\  \          */
/*           ▓█   ▓██▒░██████▒▒██▒ ░  ░         \ \__\         */
/*           ▒▒   ▓▒█░░ ▒░▓  ░▒▓▒░ ░  ░          \|__|         */
/*            ▒   ▒▒ ░░ ░ ▒  ░░▒ ░                             */
/*         F a░a s▒e && ░S░c h░ä f e r  \                      */
/*                ░  ░    ░  ░            S o  2 0 1 9         */


// TODO BRIDGE_LEN and visualisation as simulation params
// TODO give show_world a break and join it
// TODO seperate visualization from wolrd
// TODO NCURSES BRIDGE

/*▄▄▄▄▄▄▄▄▄▄▄▄▄▄▄▄▄▄▄▄▄▄▄▄▄▄ INCLUDE ▄▄▄▄▄▄▄▄▄▄▄▄▄▄▄▄▄▄▄▄▄▄▄▄▄▄*/
#include "world.h"                   // typedef of side
/*▀▀▀▀▀▀▀▀▀▀▀▀▀▀▀▀▀▀▀▀▀▀▀▀▀▀ INCLUDE ▀▀▀▀▀▀▀▀▀▀▀▀▀▀▀▀▀▀▀▀▀▀▀▀▀▀*/



/*▄▄▄▄▄▄▄▄▄▄▄▄▄▄▄▄▄▄▄▄▄▄▄▄▄▄ DEFINE  ▄▄▄▄▄▄▄▄▄▄▄▄▄▄▄▄▄▄▄▄▄▄▄▄▄▄*/
#define BRIDGE_LEN 5
/*▀▀▀▀▀▀▀▀▀▀▀▀▀▀▀▀▀▀▀▀▀▀▀▀▀▀ DEFINE  ▀▀▀▀▀▀▀▀▀▀▀▀▀▀▀▀▀▀▀▀▀▀▀▀▀▀*/



/*▄▄▄▄▄▄▄▄▄▄▄▄▄▄▄▄▄▄▄▄▄▄▄▄▄▄ GLOBAL  ▄▄▄▄▄▄▄▄▄▄▄▄▄▄▄▄▄▄▄▄▄▄▄▄▄▄*/
uint64_t FSALP42019_WORLD_BRIDGE[BRIDGE_LEN]; // > 0 implies crash
size_t COLLISION;                             // count collisions
size_t RUN_MAX;
/*▀▀▀▀▀▀▀▀▀▀▀▀▀▀▀▀▀▀▀▀▀▀▀▀▀▀ GLOBAL  ▀▀▀▀▀▀▀▀▀▀▀▀▀▀▀▀▀▀▀▀▀▀▀▀▀▀*/



/*▄▄▄▄▄▄▄▄▄▄▄▄▄▄▄▄▄▄▄▄▄▄▄▄▄▄ STRUCTS ▄▄▄▄▄▄▄▄▄▄▄▄▄▄▄▄▄▄▄▄▄▄▄▄▄▄*/
typedef struct {
  uint64_t * plate;
  side side;
  uint64_t pos;
  char algorithm;
} carArgs;

typedef struct {
  pthread_t car;
  int retv;
  carArgs args;
} pCar;
/*▀▀▀▀▀▀▀▀▀▀▀▀▀▀▀▀▀▀▀▀▀▀▀▀▀▀ STRUCTS ▀▀▀▀▀▀▀▀▀▀▀▀▀▀▀▀▀▀▀▀▀▀▀▀▀▀*/



/*▄▄▄▄▄▄▄▄▄▄▄▄▄▄▄▄▄▄▄▄▄▄▄ IMPLEMENTATION ▄▄▄▄▄▄▄▄▄▄▄▄▄▄▄▄▄▄▄▄▄▄*/

/***************************************************************/
/* function run_world :: runs the world                        */
void run_world(worldParameter * param)
{
  pthread_t        gfx;
  pCar             cars[param->num_of_obj];
  uint64_t         numbers[param->num_of_obj];
  for (size_t i = 0; i < param->num_of_obj; i++)                 // give the cars ascending ids
  {
    numbers[i]=i;
  }
  RUN_MAX = param->num_of_runs;
  srand(time(NULL));                                             // init random seed

  COLLISION = 0;                                                 // init collision counter
  for (size_t i = 0; i < BRIDGE_LEN; i++)                        // init empty bridge
    FSALP42019_WORLD_BRIDGE[i] = 0;

  int retshow = 0;
  if ( 0 != (retshow = pthread_create(&gfx, NULL, show_world, NULL) ))
    exit(EXIT_FAILURE);

  for (size_t i = 0; i < param->num_of_obj; i++)                 // spawn the cars
  {
    cars[i] = (pCar) {
      .args.plate = &numbers[i],
      .args.side = (side) (rand() % 2),
      .args.algorithm = param->algorithm
    };
    cars[i].retv = pthread_create(&cars[i].car, NULL, drive_car,
                                  (void*) &cars[i].args);
    if ( cars[i].retv )                                          // use retv for error-handling
      printf("Car Number %ld is not functional (ErrorCode: %d)\n",
             i, cars[i].retv);
  }

  for (size_t i = 0; i < param->num_of_obj; i++)                 // ensure cars are done driving
    pthread_join(cars[i].car, NULL);

  // pthread_join(retshow, NULL);
  bridge_status();
  printf("there were %ld collisions\n", COLLISION);
  exit(0);
}

/***************************************************************/
/* function drive_car :: provides cars to pthread_create       */
void * drive_car(void * args)
{
  struct timespec sleep_timer, sleep_remainder;                  // to stay on bridge
  sleep_timer.tv_nsec = (rand() % 10) * 100000000L;              // 10th - 100th second
  sleep_timer.tv_sec  = 0;
  carArgs * cargs = (carArgs *) args;                            // retriving arguments
  uint64_t * number = cargs->plate;
  char  algorithm = cargs->algorithm;
  pthread_t thread_id = pthread_self();
  print_thread_id(thread_id, *number);


  for (size_t r = 0 ; r < RUN_MAX; r++)
  {
        switch (algorithm)
        {
          case 'n':
            for (size_t i = 0; i < BRIDGE_LEN; i++)
            {
              drive(i,&sleep_timer,&sleep_remainder);
            }
            break;
          case 'l':
            lock_lamport(*number);
            for (size_t i = 0; i < BRIDGE_LEN; i++)
            {
              drive(i,&sleep_timer,&sleep_remainder);
            }
            unlock_lamport(*number);
            break;
          case 'd':
            lock_dekker(* number);
            for (size_t i = 0; i < BRIDGE_LEN; i++)
            {
              drive(i,&sleep_timer,&sleep_remainder);
            }
            unlock_dekker(* number);
            break;
          case 'p':
            lock_peterson(* number);
            for (size_t i = 0; i < BRIDGE_LEN; i++)
            {
              drive(i,&sleep_timer,&sleep_remainder);
            }
            unlock_peterson(* number);
            break;
          case 'm':
            pthread_mutex_lock(&m_lock);
            for (size_t i = 0; i < BRIDGE_LEN; i++)
            {
              drive(i,&sleep_timer,&sleep_remainder);
            }
            pthread_mutex_unlock(&m_lock);
            break;
        }
        /*********/
  }
  return NULL;
}

/***************************************************************/
/* helper function to drive car */
void drive(size_t i,
           struct timespec * sleep_timer,
           struct timespec * sleep_remainder)
{
  ++FSALP42019_WORLD_BRIDGE[i];
  nanosleep(sleep_timer, sleep_remainder);                       // as above, to make the car stay on bridge
  bridge_status();
  --FSALP42019_WORLD_BRIDGE[i];
}

/***************************************************************/
/* Checks for collisions                                       */
int64_t bridge_status(void)
{
  for (size_t i = 0; i < BRIDGE_LEN; i++)
  {
    uint64_t tmp = FSALP42019_WORLD_BRIDGE[i];
    // check for collision
    if (tmp > 1)
      ++COLLISION;
  }
  return 0;
}

/***************************************************************/
/* Checks for collisions                                       */
void * show_world(void * arg)
{
  (void) arg; // UNUSED
  struct timespec sleep_timer, sleep_remainder;
  sleep_timer.tv_nsec = 1000000L;                                // 10th - 100th second
  sleep_timer.tv_sec  = 0;

  while (1)
  {
    print_bridge();
    nanosleep(&sleep_timer, &sleep_remainder);
  }
  return NULL;
}

/***************************************************************/
/* function print_thread_id :: helper function to print id     */
/* human readable thread safe                                  */
void print_thread_id(pthread_t thread_id, uint64_t number)
{
  unsigned char * pthread_id = (unsigned char*)(void*)(&thread_id);  // get pointer to the pthread id struct
  uint64_t buffer_position        = 0;
  char buffer[sizeof(thread_id)+40];                             // 1 byte = 2 hex chars

  buffer_position  = sprintf(buffer+buffer_position,
                             "Car %"PRIX64" pthread id 0x", number);
  for (size_t i = 0; i < sizeof(thread_id); i++)
    buffer_position += sprintf(buffer+buffer_position, "%02x",
                               (unsigned)(pthread_id[i]));
  buffer_position += sprintf(buffer+buffer_position, " rolling.\n");
  buffer[buffer_position] = 0;
  printf("%s", buffer);
}

/***************************************************************/
/* print bridge as simple c string to stdout                   */
void print_bridge(void)
{
  uint64_t buffer_position = 0;
  char buffer[BRIDGE_LEN*30];

  buffer_position  = sprintf(buffer+buffer_position, "=");
  for (size_t i = 0; i < BRIDGE_LEN; i++)
  {
    uint64_t tmp = FSALP42019_WORLD_BRIDGE[i];
    // check for collision
    if (tmp > 1)
    {
      buffer_position += sprintf(buffer+buffer_position, "x=");
    }
    else
    {
      buffer_position += sprintf(buffer+buffer_position,
                                 "%"PRIx64"=", tmp);
    }
  }

  buffer_position += sprintf(buffer+buffer_position, "\n");
  buffer[buffer_position] = 0;
  printf("%s", buffer);
}

/*▀▀▀▀▀▀▀▀▀▀▀▀▀▀▀▀▀▀▀▀▀▀ IMPLEMENTATION ▀▀▀▀▀▀▀▀▀▀▀▀▀▀▀▀▀▀▀▀▀▀▀*/
