/*                                     ___   ___               */
/*           ▄▄▄       ██▓     ██▓███  |\  \ |\  \             */
/*          ▒████▄    ▓██▒    ▓██░  ██▒\ \  \\_\  \            */
/*          ▒██  ▀█▄  ▒██░    ▓██░ ██▓▒ \ \______  \           */
/*          ░██▄▄▄▄██ ▒██░    ▒██▄█▓▒ ▒  \|_____|\  \          */
/*           ▓█   ▓██▒░██████▒▒██▒ ░  ░         \ \__\         */
/*           ▒▒   ▓▒█░░ ▒░▓  ░▒▓▒░ ░  ░          \|__|         */
/*            ▒   ▒▒ ░░ ░ ▒  ░░▒ ░                             */
/*         F a░a s▒e && ░S░c h░ä f e r  \                      */
/*                ░  ░    ░  ░            S o  2 0 1 9         */



/* Usage: <program> [-h] [-l <lock>] [-n <#cars>] [-r <#uns>]  */
/* Where lock is one of: lamport, dekker, peterson, none       */



/*▄▄▄▄▄▄▄▄▄▄▄▄▄▄▄▄▄▄▄▄▄▄▄▄▄▄ INCLUDE ▄▄▄▄▄▄▄▄▄▄▄▄▄▄▄▄▄▄▄▄▄▄▄▄▄▄*/
#include "simulation.h"
#include "libs/globals.h"
/*▀▀▀▀▀▀▀▀▀▀▀▀▀▀▀▀▀▀▀▀▀▀▀▀▀▀ INCLUDE ▀▀▀▀▀▀▀▀▀▀▀▀▀▀▀▀▀▀▀▀▀▀▀▀▀▀*/



/*▄▄▄▄▄▄▄▄▄▄▄▄▄▄▄▄▄▄▄▄▄▄▄▄▄▄ MAIN    ▄▄▄▄▄▄▄▄▄▄▄▄▄▄▄▄▄▄▄▄▄▄▄▄▄▄*/
int main(int argc, char *argv[])
{
  int64_t help_is_set = 0;
  worldParameter parameter = {.algorithm = 'n', .num_of_obj = 2,
                              .num_of_runs = 2};

  /* parsing commandline arguments */
  int64_t opt;
  while ( -1 != (opt = getopt(argc, argv, "hl:n:r:")) )
  {
    switch (opt) {
      case 'h':
        printhelp(stdout, argv[0]);
        help_is_set = 1;
        break;
      case 'l':
        /* use the first letter of an algorithm as its unique */
        if ('l' == optarg[0] || 'p' == optarg[0] ||
            'd' == optarg[0] || 'n' == optarg[0] || 'm' == optarg[0])
        {
          parameter.algorithm = optarg[0];
          break;
        }
        else
        {
          printhelp(stderr, argv[0]);
          exit(EXIT_FAILURE);
        }
      case 'n':
        parameter.num_of_obj = atoi(optarg); // TODO use strtol
        break;
      case 'r':
        parameter.num_of_runs = atoi(optarg);
        break;
      default: /* .?. */
        printhelp(stderr, argv[0]);
        exit(EXIT_FAILURE);
    }
  }

  /* check for inconsistend parameter */
  if ( 2 != parameter.num_of_obj
      && ( 'd' == parameter.algorithm || 'p' == parameter.algorithm ) )
    parameter.num_of_obj = 2; // bold move to know it better

  /* if no help shall be displayed, start the world */
  if (!help_is_set)
  {
    /* pthread_mutex are staticly initilased */
    if ( 'm' != parameter.algorithm )
    {
      /* setup memory for the locks */
      locks_init(parameter.num_of_obj);
    }
    run_world(&parameter);
    /* free memory */
    locks_free();
  }

  exit(EXIT_SUCCESS);
}
/*▀▀▀▀▀▀▀▀▀▀▀▀▀▀▀▀▀▀▀▀▀▀▀▀▀▀ MAIN    ▀▀▀▀▀▀▀▀▀▀▀▀▀▀▀▀▀▀▀▀▀▀▀▀▀▀*/



/*▄▄▄▄▄▄▄▄▄▄▄▄▄▄▄▄▄▄▄▄▄▄ HELPER FUNCTIONS ▄▄▄▄▄▄▄▄▄▄▄▄▄▄▄▄▄▄▄▄▄*/
void printhelp(FILE * where, char * what)
{
  fprintf(where,
      "Usage: %s [-h] [-l <lockname>] [-n <num cars>] [-r <num_runs>] \n \
      lockname: lamport, dekker, peterson, none\n \
      only the first charater will be considered\n",
      what);
}
/*▀▀▀▀▀▀▀▀▀▀▀▀▀▀▀▀▀▀▀▀▀▀▀▀ HELPER FUNCTIONS ▀▀▀▀▀▀▀▀▀▀▀▀▀▀▀▀▀▀▀▀▀▀▀▀*/
